/*
 * ex10.c : a simple and ioctl controlled multi loopback char
 *          device driver with semaphores and __kernel timer__
 *
 * Copyright (C) 2001 Julien Gaulmin <julien.gaulmin@fr.alcove.com>, Alcve
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA or look at http://www.gnu.org/copyleft/gpl.html
 */


/*
 * Includes
 */

#include <linux/kernel.h>	/* printk() */
#include <linux/module.h>	/* modules */
#include <linux/init.h>		/* module_{init,exit}() */
#include <linux/devfs_fs_kernel.h>	/* devfs support */
#include <linux/malloc.h>	/* kmalloc()/kfree() */
#include <asm/uaccess.h>	/* copy_{from,to}_user() */

#include "ex.h"


/*
 * Arguments
 */

static int major = 0; /* Major number */
static int buf_size = DEF_BUF_SIZE; /* Buffer size */
static int max_devices = MAX_DEVICES; /* Maximum supported devices */

MODULE_DESCRIPTION("Simple and ioctl controlled multi loopback char device driver with semaphores and __kernel timer__ (ex10)");
MODULE_AUTHOR("Julien Gaulmin, Alcove");
MODULE_SUPPORTED_DEVICE("Nothing in particular");
MODULE_PARM_DESC(major, "Static major number (none = dynamic)");
MODULE_PARM_DESC(buf_size, "Buffer size");
MODULE_PARM_DESC(max_devices, "Maximum supported devices");
MODULE_PARM(major, "i");
MODULE_PARM(buf_size, "i");
MODULE_PARM(max_devices, "i");


/*
 * Global variables
 */

static struct struct_private {
	devfs_handle_t devfs_handle; /* Devfs file handle */
	char *buffer; /* The buffer */
	int num; /* Number of available bytes in the buffer */
	int blk_mode; /* Blocking mode */
	struct semaphore buffer_mutex; /* Protect concurrent read accesses */
	int end_timer; /* Timer expiration flag */
	struct timer_list timer; /* The timer */
} *ex10_privates;


/*
 * Event handlers
 */

static void timer_handler(unsigned long arg) {
	/* Mark the timer expiration flag */
	ex10_privates[arg].end_timer = 1;

	/* Wake up one blocked process (buggy as all should be woken up) */
	up(&ex10_privates[arg].buffer_mutex);
}


/*
 * File operations
 */

static ssize_t ex10_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
	int real;
	int minor = MINOR(file->f_dentry->d_inode->i_rdev);

	/* Sleep if no data available and the driver is in blocking mode.
	 * Only one reader is woken up at a time. */
	if (ex10_privates[minor].blk_mode == RD_BLK) {
		mod_timer(&ex10_privates[minor].timer, jiffies + 10 * HZ);

		if (down_interruptible(&ex10_privates[minor].buffer_mutex)) {
			printk(KERN_DEBUG "ex10: woke up by signal\n");

			/* Cancel the timer */
			del_timer(&ex10_privates[minor].timer);

			return -ERESTARTSYS;
		}

		if (ex10_privates[minor].end_timer != 0) {
			printk(KERN_DEBUG "ex10: woke up by timer\n");

			ex10_privates[minor].end_timer = 0;

			return -ETIME;
		} else {
			printk(KERN_DEBUG "ex10: woke up by writer\n");

			/* Cancel the timer */
			del_timer(&ex10_privates[minor].timer);
		}
	}

	if (count <= ex10_privates[minor].num)
		real = count;
	else
		real = ex10_privates[minor].num;

	if (real)
		copy_to_user(buf, ex10_privates[minor].buffer, real);

	ex10_privates[minor].num = 0; /* Destructive read (no more data after a read) */

	printk(KERN_DEBUG "ex10: read %d/%d chars\n", real, count);

	return real;
}


static ssize_t ex10_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	int real;
	int minor = MINOR(file->f_dentry->d_inode->i_rdev);

	if (count <= buf_size)
		real = count;
	else
		real = buf_size;

	if (real)
		copy_from_user(ex10_privates[minor].buffer, buf, real);

	ex10_privates[minor].num = real; /* Destructive write (overwrite previous data if any) */

	/* Cancel the timer if not expired */
	if (ex10_privates[minor].end_timer == 0)
		del_timer(&ex10_privates[minor].timer);

	/* Wake up one blocked process */
	up(&ex10_privates[minor].buffer_mutex);

	printk(KERN_DEBUG "ex10: writed %d/%d chars\n", real, count);

	return real;
}


static int ex10_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	int minor = MINOR(inode->i_rdev);

	printk(KERN_DEBUG "ex10: ioctl(), %d users\n", MOD_IN_USE);

	switch (cmd) {
	case EX_MODE :
		switch (arg) {
		case RD_BLK :
			ex10_privates[minor].blk_mode = RD_BLK;
			break;

		case RD_NOBLK :
			ex10_privates[minor].blk_mode = RD_NOBLK;
			break;

		default :
			printk(KERN_WARNING "ex10: arg %x unsupported in the EX_MODE ioctl command\n", (int)arg);

			return -EINVAL;
		}
		break;

	default :
		printk(KERN_WARNING "ex10: 0x%x unsupported ioctl command\n", cmd);

		return -EINVAL;
	}

	return 0;
}


static int ex10_open(struct inode *inode, struct file *file)
{
	int minor = MINOR(inode->i_rdev);

	if (minor >= max_devices)
		return -ENODEV;

	printk(KERN_DEBUG "ex10: open(), %d users\n", MOD_IN_USE);

	return 0;
}


static int ex10_release(struct inode *inode, struct file *file)
{
	int minor = MINOR(inode->i_rdev);

	if (minor >= max_devices)
		return -ENODEV;

	printk(KERN_DEBUG "ex10: close(), %d users\n", MOD_IN_USE);

	return 0;
}


static struct file_operations ex10_fops = {
	owner:		THIS_MODULE,
	read:		ex10_read,
	write:		ex10_write,
	ioctl:		ex10_ioctl,
	open:		ex10_open,
	release:	ex10_release,
};


/*
 * Init and Exit
 */

static int __init ex10_init(void)
{
	int i, j, ret;

	/* Register the device driver */
	ret = devfs_register_chrdev(major, "ex10", &ex10_fops);
	if (ret < 0) {
		printk(KERN_WARNING "ex10: unable to get a major\n");

		return ret;
	}

	if (major == 0)
		major = ret; /* dynamic value */

	/* Allocate private structures */
	ex10_privates = (struct struct_private *)kmalloc(max_devices * sizeof(struct struct_private), GFP_KERNEL);
	if (ex10_privates != NULL) {
		printk(KERN_DEBUG "ex10: allocated a %dx%d bytes for private structures\n", max_devices, sizeof(struct struct_private));
	} else {
		printk(KERN_DEBUG "ex10: unable to allocate a %dx%d bytes for private structures\n", max_devices, sizeof(struct struct_private));

		devfs_unregister_chrdev(major, "ex10");

		return -ENOMEM;
	}

	/* Setup private structures */
	for (i = 0; i < max_devices; i++) {
		char name[20];
		
		/* Create the devfs entry */
		sprintf(name, "ex10_%d", i);
		
		ex10_privates[i].devfs_handle = devfs_register(NULL, name, DEVFS_FL_DEFAULT, major, i, S_IFCHR | S_IRUGO | S_IWUSR, &ex10_fops, &ex10_privates[i]);

		/* Allocate the buffer */
		ex10_privates[i].buffer = (char *)kmalloc(buf_size, GFP_KERNEL);
		if (ex10_privates[i].buffer != NULL) {
			printk(KERN_DEBUG "ex10: allocated a %d bytes buffer for device %d\n", buf_size, i);
		} else {
			printk(KERN_WARNING "ex10: unable to allocate a %d bytes buffer for device %d\n", buf_size, i);

			i--; /* Current index doesn't need cleanup */
			ret = -ENOMEM;
			goto cleanup;
		}

		/* Init values */
		ex10_privates[i].num = 0;
		ex10_privates[i].blk_mode = RD_NOBLK;
		ex10_privates[i].end_timer = 0; 

		/* Initialize the semaphore */
		init_MUTEX_LOCKED(&ex10_privates[i].buffer_mutex);

		/* Initialize the timer */
		init_timer(&ex10_privates[i].timer);
		ex10_privates[i].timer.function = timer_handler;
		ex10_privates[i].timer.data = i; /* Pass the minor has handler parameter */
	}

	printk(KERN_INFO "ex10: successfully loaded with major %d\n", major);

	return 0;

cleanup:
	for (j = i; j >= 0; j--) {
		devfs_unregister(ex10_privates[j].devfs_handle);

		kfree(ex10_privates[j].buffer);
	}

	kfree(ex10_privates);

	devfs_unregister_chrdev(major, "ex10");

	return ret;
}


static void __exit ex10_exit(void)
{
	int ret, i;

	ret = devfs_unregister_chrdev(major, "ex10");
	if (ret < 0) {
		printk(KERN_WARNING "ex10: error while unregistering\n");

		return;
	}

	for (i = 0; i < max_devices; i++) {
		devfs_unregister(ex10_privates[i].devfs_handle);

		kfree(ex10_privates[i].buffer);
	}

	kfree(ex10_privates);

	printk(KERN_INFO "ex10: successfully unloaded\n");
}


/*
 * Module entry points
 */

module_init(ex10_init);
module_exit(ex10_exit);
