/*
 * ex5.c : a simple __loopback__ char device driver
 *
 * Copyright (C) 2001 Julien Gaulmin <julien.gaulmin@fr.alcove.com>, Alcve
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA or look at http://www.gnu.org/copyleft/gpl.html
 */


/*
 * Includes
 */

#include <linux/kernel.h>	/* printk() */
#include <linux/module.h>	/* modules */
#include <linux/init.h>		/* module_{init,exit}() */
#include <linux/devfs_fs_kernel.h>	/* devfs support */
#include <linux/malloc.h>	/* kmalloc()/kfree() */
#include <asm/uaccess.h>	/* copy_{from,to}_user() */

#include "ex.h"


/*
 * Arguments
 */

static int major = 0; /* Major number */
static int buf_size = DEF_BUF_SIZE; /* Buffer size */

MODULE_DESCRIPTION("Simple __loopback__ char device driver (ex5)");
MODULE_AUTHOR("Julien Gaulmin, Alcove");
MODULE_SUPPORTED_DEVICE("Nothing in particular");
MODULE_PARM_DESC(major, "Static major number (none = dynamic)");
MODULE_PARM_DESC(buf_size, "Buffer size");
MODULE_PARM(major, "i");
MODULE_PARM(buf_size, "i");


/*
 * Global variables
 */

static devfs_handle_t devfs_handle; /* Devfs file handle */
static char *buffer; /* The buffer */
static int num = 0; /* Number of available bytes in the buffer */


/*
 * File operations
 */

static ssize_t ex5_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
	int real;

	if (count <= num)
		real = count;
	else
		real = num;

	if (real)
		copy_to_user(buf, buffer, real);

	num = 0; /* Destructive read (no more data after a read) */

	printk(KERN_DEBUG "ex5: read %d/%d chars\n", real, count);

	return real;
}


static ssize_t ex5_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	int real;

	if (count <= buf_size)
		real = count;
	else
		real = buf_size;

	if (real)
		copy_from_user(buffer, buf, real);

	num = real; /* Destructive write (overwrite previous data if any) */

	printk(KERN_DEBUG "ex5: writed %d/%d chars\n", real, count);

	return real;
}


static int ex5_open(struct inode *inode, struct file *file)
{
	printk(KERN_DEBUG "ex5: open(), %d users\n", MOD_IN_USE);

	return 0;
}


static int ex5_release(struct inode *inode, struct file *file)
{
	printk(KERN_DEBUG "ex5: close(), %d users\n", MOD_IN_USE);

	return 0;
}


static struct file_operations ex5_fops = {
	owner:		THIS_MODULE,
	read:		ex5_read,
	write:		ex5_write,
	open:		ex5_open,
	release:	ex5_release,
};


/*
 * Init and Exit
 */

static int __init ex5_init(void)
{
	int ret;

	ret = devfs_register_chrdev(major, "ex5", &ex5_fops);
	if (ret < 0) {
		printk(KERN_WARNING "ex5: unable to get a major\n");

		return ret;
	}

	if (major == 0)
		major = ret; /* dynamic value */

	devfs_handle = devfs_register(NULL, "ex5", DEVFS_FL_DEFAULT, major, 0, S_IFCHR | S_IRUGO | S_IWUSR, &ex5_fops, NULL);

	buffer = (char *)kmalloc(buf_size, GFP_KERNEL);
	if (buffer != NULL) {
		printk(KERN_DEBUG "ex5: allocated a %d bytes buffer\n", buf_size);
	} else {
		printk(KERN_WARNING "ex5: unable to allocate a %d bytes buffer\n", buf_size);

		devfs_unregister_chrdev(major, "ex5");

		return -ENOMEM;
	}

	printk(KERN_INFO "ex5: successfully loaded with major %d\n", major);

	return 0;
}


static void __exit ex5_exit(void)
{
	int ret;

	ret = devfs_unregister_chrdev(major, "ex5");
	if (ret < 0) {
		printk(KERN_WARNING "ex5: error while unregistering\n");

		return;
	}

	devfs_unregister(devfs_handle);

	kfree(buffer);

	printk(KERN_INFO "ex5: successfully unloaded\n");
}


/*
 * Module entry points
 */

module_init(ex5_init);
module_exit(ex5_exit);
