/*
 * ex8.c : a simple ioctl controlled __multi__ loopback char device driver
 *
 * Copyright (C) 2001 Julien Gaulmin <julien.gaulmin@fr.alcove.com>, Alcve
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA or look at http://www.gnu.org/copyleft/gpl.html
 */


/*
 * Includes
 */

#include <linux/kernel.h>	/* printk() */
#include <linux/module.h>	/* modules */
#include <linux/init.h>		/* module_{init,exit}() */
#include <linux/devfs_fs_kernel.h>	/* devfs support */
#include <linux/malloc.h>	/* kmalloc()/kfree() */
#include <asm/uaccess.h>	/* copy_{from,to}_user() */

#include "ex.h"


/*
 * Arguments
 */

static int major = 0; /* Major number */
static int buf_size = DEF_BUF_SIZE; /* Buffer size */
static int max_devices = MAX_DEVICES; /* Maximum supported devices */

MODULE_DESCRIPTION("Simple ioctl controlled __multi__ loopback char device driver (ex8)");
MODULE_AUTHOR("Julien Gaulmin, Alcove");
MODULE_SUPPORTED_DEVICE("Nothing in particular");
MODULE_PARM_DESC(major, "Static major number (none = dynamic)");
MODULE_PARM_DESC(buf_size, "Buffer size");
MODULE_PARM_DESC(max_devices, "Maximum supported devices");
MODULE_PARM(major, "i");
MODULE_PARM(buf_size, "i");
MODULE_PARM(max_devices, "i");


/*
 * Global variables
 */

static struct struct_private {
	devfs_handle_t devfs_handle; /* Devfs file handle */
	char *buffer; /* The buffer */
	int num; /* Number of available bytes in the buffer */
	wait_queue_head_t read_wait_queue; /* Read wait queue */
	int blk_mode; /* Blocking mode */
} *ex8_privates;


/*
 * File operations
 */

static ssize_t ex8_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
	int real;
	int minor = MINOR(file->f_dentry->d_inode->i_rdev);

	/* Sleep if no data available and the driver is in blocking mode.
	 * Everybody is woken up but only one can read the data (destructive
	 * mode). */
	while (ex8_privates[minor].num == 0 && ex8_privates[minor].blk_mode == RD_BLK) {
		interruptible_sleep_on(&ex8_privates[minor].read_wait_queue);

		if (signal_pending(current)) {
			printk(KERN_DEBUG "ex8: woke up by signal\n");

			return -ERESTARTSYS;
		} 
	}

	if (count <= ex8_privates[minor].num)
		real = count;
	else
		real = ex8_privates[minor].num;

	if (real)
		copy_to_user(buf, ex8_privates[minor].buffer, real);

	ex8_privates[minor].num = 0; /* Destructive read (no more data after a read) */

	printk(KERN_DEBUG "ex8: read %d/%d chars\n", real, count);

	return real;
}


static ssize_t ex8_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	int real;
	int minor = MINOR(file->f_dentry->d_inode->i_rdev);

	if (count <= buf_size)
		real = count;
	else
		real = buf_size;

	if (real)
		copy_from_user(ex8_privates[minor].buffer, buf, real);

	ex8_privates[minor].num = real; /* Destructive write (overwrite previous data if any) */

	/* Wake up all blocked processes in read_wait_queue */
  	wake_up_interruptible(&ex8_privates[minor].read_wait_queue);

	printk(KERN_DEBUG "ex8: writed %d/%d chars\n", real, count);

	return real;
}


static int ex8_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	int minor = MINOR(inode->i_rdev);

	printk(KERN_DEBUG "ex8: ioctl(), %d users\n", MOD_IN_USE);

	switch (cmd) {
	case EX_MODE :
		switch (arg) {
		case RD_BLK :
			ex8_privates[minor].blk_mode = RD_BLK;
			break;

		case RD_NOBLK :
			ex8_privates[minor].blk_mode = RD_NOBLK;
			break;

		default :
			printk(KERN_WARNING "ex8: arg %x unsupported in the EX_MODE ioctl command\n", (int)arg);

			return -EINVAL;
		}
		break;

	default :
		printk(KERN_WARNING "ex8: 0x%x unsupported ioctl command\n", cmd);

		return -EINVAL;
	}

	return 0;
}


static int ex8_open(struct inode *inode, struct file *file)
{
	int minor = MINOR(inode->i_rdev);

	if (minor >= max_devices)
		return -ENODEV;

	printk(KERN_DEBUG "ex8: open(), %d users\n", MOD_IN_USE);

	return 0;
}


static int ex8_release(struct inode *inode, struct file *file)
{
	int minor = MINOR(inode->i_rdev);

	if (minor >= max_devices)
		return -ENODEV;

	printk(KERN_DEBUG "ex8: close(), %d users\n", MOD_IN_USE);

	return 0;
}


static struct file_operations ex8_fops = {
	owner:		THIS_MODULE,
	read:		ex8_read,
	write:		ex8_write,
	ioctl:		ex8_ioctl,
	open:		ex8_open,
	release:	ex8_release,
};


/*
 * Init and Exit
 */

static int __init ex8_init(void)
{
	int i, j, ret;

	/* Register the device driver */
	ret = devfs_register_chrdev(major, "ex8", &ex8_fops);
	if (ret < 0) {
		printk(KERN_WARNING "ex8: unable to get a major\n");

		return ret;
	}

	if (major == 0)
		major = ret; /* dynamic value */

	/* Allocate private structures */
	ex8_privates = (struct struct_private *)kmalloc(max_devices * sizeof(struct struct_private), GFP_KERNEL);
	if (ex8_privates != NULL) {
		printk(KERN_DEBUG "ex8: allocated a %dx%d bytes for private structures\n", max_devices, sizeof(struct struct_private));
	} else {
		printk(KERN_DEBUG "ex8: unable to allocate a %dx%d bytes for private structures\n", max_devices, sizeof(struct struct_private));

		devfs_unregister_chrdev(major, "ex8");

		return -ENOMEM;
	}

	/* Setup private structures */
	for (i = 0; i < max_devices; i++) {
		char name[20];
		
		/* Create the devfs entry */
		sprintf(name, "ex8_%d", i);
		
		ex8_privates[i].devfs_handle = devfs_register(NULL, name, DEVFS_FL_DEFAULT, major, i, S_IFCHR | S_IRUGO | S_IWUSR, &ex8_fops, &ex8_privates[i]);

		/* Allocate the buffer */
		ex8_privates[i].buffer = (char *)kmalloc(buf_size, GFP_KERNEL);
		if (ex8_privates[i].buffer != NULL) {
			printk(KERN_DEBUG "ex8: allocated a %d bytes buffer for device %d\n", buf_size, i);
		} else {
			printk(KERN_WARNING "ex8: unable to allocate a %d bytes buffer for device %d\n", buf_size, i);

			i--; /* Current index doesn't need cleanup */
			ret = -ENOMEM;
			goto cleanup;
		}

		/* Init values */
		ex8_privates[i].num = 0;
		init_waitqueue_head(&ex8_privates[i].read_wait_queue);
		ex8_privates[i].blk_mode = RD_NOBLK;
	}

	printk(KERN_INFO "ex8: successfully loaded with major %d\n", major);

	return 0;

cleanup:
	for (j = i; j >= 0; j--) {
		devfs_unregister(ex8_privates[j].devfs_handle);

		kfree(ex8_privates[j].buffer);
	}

	kfree(ex8_privates);

	devfs_unregister_chrdev(major, "ex8");

	return ret;
}


static void __exit ex8_exit(void)
{
	int ret, i;

	ret = devfs_unregister_chrdev(major, "ex8");
	if (ret < 0) {
		printk(KERN_WARNING "ex8: error while unregistering\n");

		return;
	}

	for (i = 0; i < max_devices; i++) {
		devfs_unregister(ex8_privates[i].devfs_handle);

		kfree(ex8_privates[i].buffer);
	}

	kfree(ex8_privates);

	printk(KERN_INFO "ex8: successfully unloaded\n");
}


/*
 * Module entry points
 */

module_init(ex8_init);
module_exit(ex8_exit);
