/*
 * genpci.c : a generic PCI driver
 *
 * Copyright (C) 2001 Julien Gaulmin <julien.gaulmin@fr.alcove.com>, Alcve
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA or look at http://www.gnu.org/copyleft/gpl.html
 */


/*
 * Includes
 */

#include <linux/kernel.h>	/* printk() */
#include <linux/module.h>	/* modules */
#include <linux/init.h>		/* module_{init,exit}() */
#include <linux/devfs_fs_kernel.h>	/* devfs support */
#include <linux/malloc.h>	/* kmalloc()/kfree() */
#include <linux/pci.h>		/* pci_module_init() */
#include <linux/list.h>		/* list_*() */
#include <asm/uaccess.h>	/* copy_{from,to}_user() */


/*
 * Arguments
 */

static int major = 0; /* Major number */

MODULE_DESCRIPTION("Generic PCI driver (genpci)");
MODULE_AUTHOR("Julien Gaulmin, Alcove");
MODULE_SUPPORTED_DEVICE("All");
MODULE_PARM_DESC(major, "Static major number (none = dynamic)");
MODULE_PARM(major, "i");


/*
 * Supported devices
 */

typedef enum {
	NC_3C905C = 0,
	NC_RT8139,
	MC_SDNS2000,
	VC_MGA1064SG,
} board_t;

static struct {
	const char *name;
} board_info[] __devinitdata = {
	{"3COM 3c905C [Tornado]"},
	{"Realtek RT8139 [Fast Ethernet]"},
	{"Sigma Designs NetStream 2000"},
	{"Matrox MGA 1064SG [Mystique]"},
};


static struct pci_device_id genpci_id_table[] __devinitdata = {
	{0x10B7, 0x9200, PCI_ANY_ID, PCI_ANY_ID, 0, 0, NC_3C905C},
	{0x10EC, 0x8139, PCI_ANY_ID, PCI_ANY_ID, 0, 0, NC_RT8139},
	{0x1105, 0x8400, PCI_ANY_ID, PCI_ANY_ID, 0, 0, MC_SDNS2000},
	{0x102B, 0x051A, PCI_ANY_ID, PCI_ANY_ID, 0, 0, VC_MGA1064SG},
	{0,}	/* 0 terminated list */
};
MODULE_DEVICE_TABLE(pci, genpci_id_table);


/*
 * Global variables
 */

static LIST_HEAD(privates);

struct struct_private {
	struct list_head	link; /* Double linked list */
	struct pci_dev	*dev; /* PCI device */
	board_t		type; /* Board type */
	int		minor; /* Minor number */
	devfs_handle_t	devfs_handle; /* Devfs file handle */
	void		*mmio[DEVICE_COUNT_RESOURCE]; /* Remaped I/O memory */
	u32		mmio_len[DEVICE_COUNT_RESOURCE]; /* Size of remaped I/O memory */
};


/*
 * Event handlers
 */

void genpci_irq_handler(int irq, void *dev_id, struct pt_regs *regs)
{
	struct struct_private *genpci_private = (struct struct_private *)dev_id;
	//struct pci_dev *dev = genpci_private->dev;

	printk(KERN_DEBUG "genpci: interrupt from device %d\n", genpci_private->minor);
}


/*
 * File operations
 */

static ssize_t genpci_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
	int i, real, bank = DEVICE_COUNT_RESOURCE;
	struct struct_private *genpci_private = file->private_data;

	/* Find the first remapped I/O memory bank to read */
	for (i = 0; i < DEVICE_COUNT_RESOURCE; i++) {
		if (genpci_private->mmio[i] != NULL) {
			bank = i;

			break;
		}
	}

	/* No bank found */
	if (bank == DEVICE_COUNT_RESOURCE) {
		printk(KERN_WARNING "genpci: no remaped I/O memory bank to read\n");

		return -ENXIO;
	}

	/* Check for overflow */
	if (count <= genpci_private->mmio_len[bank] - (int)*ppos)
		real = count;
	else
		real = genpci_private->mmio_len[bank] - (int)*ppos;

	if (real)
		copy_to_user(buf, (char*)genpci_private->mmio[bank] + (int)*ppos, real);

	printk(KERN_DEBUG "genpci: read %d/%d chars at offset %d from remaped I/O memory bank %d\n", real, count, (int)*ppos, bank);

	*ppos += real;

	return real;
}


static int genpci_open(struct inode *inode, struct file *file)
{
	int minor = MINOR(inode->i_rdev);
	struct list_head *cur;
	struct struct_private *genpci_private;

	printk(KERN_DEBUG "genpci_open()\n");

	list_for_each(cur, &privates) {
		genpci_private = list_entry(cur, struct struct_private, link);

		if (genpci_private->minor == minor) {
			file->private_data = genpci_private;
			
			return 0;
		}
	}

	printk(KERN_WARNING "genpci: minor %d not found\n", minor);
	
	return -ENODEV;
}


static int genpci_release(struct inode *inode, struct file *file)
{
	printk(KERN_DEBUG "genpci_release()\n");

	file->private_data = NULL;

	return 0;
}


static struct file_operations genpci_fops = {
	owner:		THIS_MODULE,
	read:		genpci_read,
	open:		genpci_open,
	release:	genpci_release,
};


/*
 * PCI handling
 */

static int __devinit genpci_probe(struct pci_dev *dev, const struct pci_device_id *ent)
{
	int i, ret = 0;
	char devfs_name[20];
	struct struct_private *genpci_private;

	printk(KERN_INFO "genpci: found %s\n", board_info[ent->driver_data].name);
	printk(KERN_INFO "genpci: using major %d and minor %d for this device\n", major, (int)ent->driver_data);

	/* Allocate a private structure and reference it as driver's data */
	genpci_private = (struct struct_private *)kmalloc(sizeof(struct struct_private), GFP_KERNEL);
	if (genpci_private == NULL) {
		printk(KERN_WARNING "genpci: unable to allocate private structure\n");

		ret = -ENOMEM;
		goto cleanup_kmalloc;
	}

	pci_set_drvdata(dev, genpci_private);

	/* Init private field */
	genpci_private->dev = dev;
	genpci_private->minor = genpci_private->type = ent->driver_data;

 	/* Initialize device before it's used by the driver */
	ret = pci_enable_device(dev);
	if (ret < 0) {
		printk(KERN_WARNING "genpci: unable to initialize PCI device\n");

		goto cleanup_pci_enable;
	}
	
	/* Reserve PCI I/O and memory resources */
	ret = pci_request_regions(dev, "genpci");
	if (ret < 0) {
		printk(KERN_WARNING "genpci: unable to reserve PCI resources\n");

		goto cleanup_regions;
	}

	/* Inspect PCI BARs and remap I/O memory */
	for (i=0; i < DEVICE_COUNT_RESOURCE; i++) {
		if (pci_resource_start(dev, i) != 0)
			printk(KERN_DEBUG "genpci: BAR %d (%#08x-%#08x), len=%d, flags=%#08x\n", i, (u32) pci_resource_start(dev, i), (u32) pci_resource_end(dev, i), (u32) pci_resource_len(dev, i), (u32) pci_resource_flags(dev, i));

		if (pci_resource_flags(dev, i) & IORESOURCE_MEM) {
			genpci_private->mmio[i] = ioremap(pci_resource_start(dev, i), pci_resource_len(dev, i));
			if (genpci_private->mmio[i] == NULL) {
				printk(KERN_WARNING "genpci: unable to remap I/O memory\n");

				ret = -ENOMEM;
				goto cleanup_ioremap;
			}

			genpci_private->mmio_len[i] = pci_resource_len(dev, i);
			
			printk(KERN_DEBUG "genpci: I/O memory has been remaped at %#08x\n", (u32)genpci_private->mmio[i]);
		} else {
			genpci_private->mmio[i] = NULL;
		}
	}

	/* Install the irq handler */
	ret = request_irq(dev->irq, genpci_irq_handler, SA_SHIRQ, "genpci", genpci_private);  
	if (ret < 0) {
		printk(KERN_WARNING "genpci: unable to register irq handler\n");

		goto cleanup_irq;
	}

 	/* Enable PCI bus-mastering for this device */
	pci_set_master(dev);
	
	/* Create the devfs entry */
	sprintf(devfs_name, "genpci/device_%d", ent->driver_data);
	
	genpci_private->devfs_handle = devfs_register(NULL, devfs_name, DEVFS_FL_DEFAULT, major, ent->driver_data, S_IFCHR | S_IRUGO | S_IWUSR, &genpci_fops, dev);

	/* Link the new genpci_private structure with others */
	list_add_tail(&genpci_private->link, &privates);

	return 0;

cleanup_irq:
	for (i--; i >= 0; i--) {
		if (genpci_private->mmio[i] != NULL)
			iounmap(genpci_private->mmio[i]);
	}
cleanup_ioremap:
	pci_release_regions(dev);
cleanup_regions:
	pci_disable_device(dev);
cleanup_pci_enable:
	kfree(genpci_private);
cleanup_kmalloc:
	return ret;
}


static void __devexit genpci_remove(struct pci_dev *dev)
{
	int i;
	struct struct_private *genpci_private = pci_get_drvdata(dev);

	printk(KERN_INFO "genpci: device removed\n");

	for (i=0; i < DEVICE_COUNT_RESOURCE; i++) {
		if (genpci_private->mmio[i] != NULL)
			iounmap(genpci_private->mmio[i]);
	}

	pci_release_regions(dev);

	pci_disable_device(dev);

	kfree(genpci_private);

	free_irq(dev->irq, "genpci");
	
	devfs_unregister(genpci_private->devfs_handle);

	list_del(&genpci_private->link);
}


static struct pci_driver genpci_pci_driver = {
	name:		"genpci",
	id_table:	genpci_id_table,
	probe:		genpci_probe,		/* Init one device */
	remove:		genpci_remove,		/* Remove one device */
};


/*
 * Init and Exit
 */

static int __init genpci_init(void)
{
	int ret;

	/* Register the device driver */
	ret = devfs_register_chrdev(major, "genpci", &genpci_fops);
	if (ret < 0) {
		printk(KERN_WARNING "genpci: unable to get a major\n");

		return ret;
	}

	if (major == 0)
		major = ret; /* dynamic value */

	/* Register PCI driver */
	ret = pci_module_init(&genpci_pci_driver);
	if (ret < 0) {
		printk(KERN_WARNING "genpci: unable to register PCI driver\n");

		devfs_unregister_chrdev(major, "genpci");

		return ret;
	}

	return 0;
}


static void __exit genpci_exit(void)
{
	pci_unregister_driver(&genpci_pci_driver);

	devfs_unregister_chrdev(major, "genpci");
}


/*
 * Module entry points
 */

module_init(genpci_init);
module_exit(genpci_exit);
