/*
 * ex3ter.c : a simple module with arguments and ps like in /proc updated by a
 *            __kernel thread__
 *
 * Copyright (C) 2001 Julien Gaulmin <julien.gaulmin@fr.alcove.com>, Alcve
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA or look at http://www.gnu.org/copyleft/gpl.html
 */


/*
 * Includes
 */

#include <linux/kernel.h>	/* printk() */
#include <linux/module.h>	/* modules */
#include <linux/init.h>		/* module_{init,exit}() */
#include <linux/proc_fs.h>	/* proc fs */
#include <linux/smp_lock.h>	/* {lock,unlock}_kernel() */
#include <linux/version.h>	/* KERNEL_VERSION() */

#include "ex.h"


/*
 * Macros
 */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,9)
#	define DECLARE_COMPLETION(x)	DECLARE_MUTEX_LOCKED(x)
#	define init_completion(x)
#	define complete_and_exit(x, y)	up_and_exit((x), (y))
#	define wait_for_completion(x)	down(x)
#else
#	include <linux/completion.h>
#endif


/*
 * Arguments
 */

static char *id = "ex3ter_default"; /* ID-String of the module */
static int num = 1234; /* A number */

MODULE_DESCRIPTION("Simple module with arguments and ps like in /proc updated by a __kernel thread__ (ex3ter)");
MODULE_AUTHOR("Julien Gaulmin, Alcove");
MODULE_SUPPORTED_DEVICE("Nothing in particular");
MODULE_PARM_DESC(id, "ID-String of the driver");
MODULE_PARM_DESC(num, "A number");
MODULE_PARM(id, "1-" __MODULE_STRING(MAX_ID_SIZE) "s");
MODULE_PARM(num, "i");


/*
 * Global variables
 */

static int thread_pid; /* Thread pid */
static DECLARE_COMPLETION(thread_end); /* Synchronise thread end */
static char buffer[PAGE_SIZE];
static int buffer_len;
static DECLARE_MUTEX(buffer_mutex); /* Protect the proc buffer */


/*
 * Features
 */

static int ps(char *buf) {
	struct task_struct *p;
	int len = 0;

	len += sprintf(buf + len, "Simple PS :\n");

	read_lock(&tasklist_lock);
	for_each_task(p) {
		if (len + MAX_PS_LINE_SIZE > PAGE_SIZE) {
			len += sprintf(buf + len, "[TRUNCATED]\n");

			break;
		}

		len += sprintf(buf + len, "%d\t", (int)p->pid);
		len += sprintf(buf + len, "%s\n", p->comm);
	}
	read_unlock(&tasklist_lock);

	return len;
}


int thread_core(void *arg)
{
	printk(KERN_INFO "ex3ter: kernel thread running...\n");

	/* Stand-alone kernel thread */
	lock_kernel();
	daemonize();
	strcpy(current->comm, "kex3terd");
	unlock_kernel();

	/* Receive only SIGKILL signals */
	siginitsetinv(&current->blocked, sigmask(SIGKILL));
	spin_lock(&current->sigmask_lock);
	flush_signals(current);
	spin_unlock(&current->sigmask_lock);

	/* Main loop */
	while (!signal_pending(current)) {
		/* Update the buffer */
		down(&buffer_mutex);
		buffer_len = ps(buffer);
		up(&buffer_mutex);

		/* Sleep 20 secondes */
		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(20 * HZ);

		printk(KERN_INFO "ex3ter: proc updated\n");
	}

	printk(KERN_INFO "ex3ter: kernel thread leaving...\n");

	/* Signal end and exit */
	complete_and_exit(&thread_end, 0);
}


/*
 * /proc handling
 */

static int ex3ter_get_info(char *page, char **start, off_t off, int count)
{
	int len = 0;

	/* Copy local buffer into the page allocated by the kernel */
	down(&buffer_mutex);
	strncpy(page, buffer, PAGE_SIZE);
	len = buffer_len;
	up(&buffer_mutex);

	return len;
}


/*
 * Init and Exit
 */

static int __init ex3ter_init(void)
{
	struct proc_dir_entry *proc_entry;

	printk(KERN_INFO "ex3ter initialization\n");
	printk(KERN_INFO "ex3ter arguments : id=%s, num=%d\n", id, num);

	/* create /proc entry */
	proc_entry = create_proc_entry("ex3ter", 0, NULL);
	if (proc_entry) {
		proc_entry->get_info = ex3ter_get_info;
	} else {
		printk(KERN_ERR "Can't create proc entry for ex3ter\n");

		return -EAGAIN;
	}

	/* kernel thread init */
	init_completion(&thread_end);

	thread_pid = kernel_thread(thread_core, NULL, 0);
	if (thread_pid < 0) {
		printk(KERN_ERR "Can't create kernel thread for ex3ter\n");

		return -EAGAIN;
	}

	printk(KERN_INFO "ex3ter: kernel thread launched (pid=%d)\n", thread_pid);

	return 0;
}


static void __exit ex3ter_exit(void)
{
	/* cleanup /proc entry */
	remove_proc_entry("ex3ter", NULL);

	/* kill kernel thread and wait for his death */
	kill_proc(thread_pid, SIGKILL, 1);
	wait_for_completion(&thread_end);

	printk(KERN_INFO "ex3ter exit\n");
}


/*
 * Module entry points
 */

module_init(ex3ter_init);
module_exit(ex3ter_exit);
